"""Module contains class for handling PostgreSQL DB."""
import os
import sys
import json
import logging
import psycopg2
import psycopg2.extras
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
# This line is so modules using this package as a submodule can use this.
sys.path.append(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/'))
#

class DatabaseHandler:
    def __init__(self, configHandler):
        self.configHandler = configHandler
        self.dbInfo = self.configHandler.getDbInfo()
        
        # Open cursor to the default server (named postgres)
        self.defaultConnection = self.openConnection(openDefault=True)

        if not self.dbExists(self.dbInfo["database"]):
            self.createNewDb(self.dbInfo["database"])

        # Open cursor to the server specified in the config file
        self.connection = self.openConnection()

    def openConnection(self, openDefault=False):
        """Opens a connection to DB.
        
        Parameters
        ----------
        openDefault : bool
            Open a connection to the DB named postgres, otherwise to the DB specified in config file
        """
        params = self.dbInfo.copy()
        if openDefault:
            params['database'] = 'postgres'
        logging.info("Opening connection to DB: %s", params['database'])
        connection = psycopg2.connect(**params)
        connection.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        return connection

    def closeConnection(self, connection):
        logging.info('Closing connection')
        if connection is not None:
            connection.close()

    def openCursor(self, connection):
        return connection.cursor(cursor_factory=psycopg2.extras.DictCursor)

    def closeCursor(self, cursor):
        logging.info('Closing cursor')
        try:
            cursor.close()
        except (psycopg2.DatabaseError) as error:
            logging.warning(error)

    def checkServerConnection(self):
        logging.info('Checking connection to Postgres server')
        if self.executeQuery(self.connection, 'SELECT version()').fetchone() is not None:
            logging.info('Server connection confirmed')

    def dbExists(self, dbName):
        result = None
        sqlQuery = 'SELECT datname FROM pg_catalog.pg_database WHERE datname=\'' + dbName + '\''
        if self.executeQuery(self.defaultConnection, sqlQuery).fetchone() is None:
            result = False
        else:
            result = True
        logging.debug("DB named %s existence: %s", dbName, str(result))
        return result

    def tableExists(self, tableName):
        result = None
        sqlQuery = "SELECT * FROM information_schema.tables WHERE table_name=\'" + tableName + "\';"
        if self.executeQuery(self.connection, sqlQuery).fetchone() is None:
            result = False
        else:
            result = True
        logging.info('Table %s exists: %s', tableName, str(result))
        return result

    def countRecords(self, tableName):
        """Checks the count of records in a table."""
        sqlQuery = 'SELECT COUNT(*) FROM \"' + tableName + '\";'
        return self.executeQuery(self.connection, sqlQuery).fetchone()[0]

    def dropTable(self, tableName):
        logging.info('Attempting to drop table: %s', tableName)
        self.executeQuery(self.connection, 'DROP TABLE ' + tableName + ';')
        logging.info("Dropped table: %s", tableName)

    def addTableToDb(self, tableName, columnsInfoPath, nonElementSectionName, elementSectionName):
        logging.info('Attempting to add table')

        # Open the json with the list of columns we're interested in
        with open(columnsInfoPath) as fileReader:
            columnsInfo = json.load(fileReader)
        
        # Make the SQL query
        sqlQuery = 'CREATE TABLE \"' + tableName + '\" ('

        nonElementsColumns = columnsInfo[nonElementSectionName]
        for columnName in nonElementsColumns:
            sqlQuery = sqlQuery + "\"" + columnName + '\" ' + nonElementsColumns[columnName]['db_datatype'] + " " + nonElementsColumns[columnName]['constraints'] + ','

        elementColumns = columnsInfo[elementSectionName]
        for columnName in elementColumns:
            if not elementColumns[columnName]['calculation_only']:
                sqlQuery = sqlQuery + "\"" + columnName + '\" ' + elementColumns[columnName]['db_datatype'] + ','

        sqlQuery = sqlQuery[:-1] + ');'
        self.executeQuery(self.connection, sqlQuery)
        self.tableExists(tableName)

    def createNewDb(self, dbName):
        logging.info('Attempting to create a new DB')
        self.executeQuery(self.defaultConnection, 'CREATE DATABASE \"' + dbName + '\";')
        self.dbExists(dbName)

    def dropDb(self, dbName):
        logging.info('Attempting to drop a new DB')
        self.closeConnection(self.connection)
        self.executeQuery(self.defaultConnection, 'DROP DATABASE \"' + dbName + '\";')
        self.dbExists(dbName)

    def executeQuery(self, connection, query, values=None):
        cursor = self.openCursor(connection)
        try:
            if values is None:
                cursor.execute(query)
            else:
                cursor.execute(query, values)
        except (psycopg2.DatabaseError) as error:
            logging.warning(error)
        return cursor


"""Contains script that moves all DCM tag-values from a directory of DCMs into a PostgreSQL DB."""
import os
import logging
import sys
import json
from datetime import datetime
from dateutil.relativedelta import relativedelta
import psycopg2
from pathlib import Path
import pydicom as pdm
# This line is so modules using this package as a submodule can use this.
sys.path.append(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/'))
#

class DicomToDatabase:
    def __init__(self, configHandler, dbHandler):
        self.configHandler = configHandler
        self.dbHandler = dbHandler

    def dicomToDb(self, dbName, metaTableName, columnsInfoPath):
        logging.info('Attempting to store DICOM metadata from DCMs in a folder to Postgres DB')
        
        with open(columnsInfoPath) as fileReader:
            elementsDict = json.load(fileReader)
        elementsOriginal = elementsDict["elements"]

        folderRelPath = "./" + self.configHandler.getDatasetName()
        pathlist = Path(folderRelPath).glob('**/*.dcm')
        for path in pathlist:
            elements = elementsOriginal.copy()

            filePath = str(path)
            logging.debug('Storing: ' + filePath)

            # Insert the DICOM metadata as a new record in the Postgres DB
            sqlQuery, values = self.createSqlQuery(metaTableName, elements, filePath)
            self.dbHandler.executeQuery(self.dbHandler.connection, sqlQuery, values)
                    
        logging.info('Done storing metadata')

    def createSqlQuery(self, metaTableName, elements, filePath):
        """Create the SQL query for inserting a record."""
        dcm = pdm.dcmread(filePath)
        # Go through the list of elements and try to read the value
        for elementName in elements.keys():
            tag = elements[elementName]['tag']
            [groupNum, elementNum] = tag.split(',')
            try:
                element = dcm[int(groupNum, 16), int(elementNum, 16)]
                if element.VR == 'DS':
                    if element.VM == 1:
                        value = int(float(element.value))
                    elif element.VM > 1:
                        value = int(float(element.value[0]))
                else:
                    value = element.value
                elements[elementName]['value'] = value
            except (KeyError) as tag: # if the value isn't there, then set it as None
                logging.debug('Cannot read the following DICOM tag: ' + str(tag))
                elements[elementName]['value'] = None
                continue

        # Adjust the data as necessary before storing
        self.dataAdjustments(elements)

        # Create the list of values that we're going to use to build the query
        names = ['file_name', 'file_path']
        values = [filePath.split(os.sep)[-1], filePath]
        placeholders = ['%s', '%s']
        # Append any value to this list that isn't birth_date or study_date
        for elementName in elements.keys():
            if not elements[elementName]['calculation_only']:
                names.append(elementName)
                values.append(elements[elementName]['value'])
                placeholders.append('%s')

        # Build the SQL query
        sqlQuery = 'INSERT INTO ' + metaTableName + ' (' + ', '.join(names) + ')' + os.linesep + 'VALUES (' + ', '.join(placeholders) + ');'

        return (sqlQuery, values)

    def dataAdjustments(self, elements):
        """Make adjustments to the data if they are needed."""
        # Calculate patient_age if it's in the list of elements but wasn't read from the DCM
        if ('patient_age' in elements) and (elements['patient_age']['value'] is None):
            if (elements['patient_birth_date']['value'] is not None) and (elements['study_date']['value'] is not None):
                logging.info('Calculating the patient age (0x0010, 0x1010)')
                patientBirthDate = elements['patient_birth_date']['value']
                studyDate = elements['study_date']['value']

                patientBirthDate = datetime.strptime(patientBirthDate, '%Y%m%d')
                studyDate = datetime.strptime(studyDate, '%Y%m%d')
                elements['patient_age']['value'] = relativedelta(studyDate, patientBirthDate).years

        # Combine patient_orientation into a string, otherwise it is treated as a list
        if ('patient_orientation' in elements) and (elements['patient_orientation']['value'] is not None):
            elements['patient_orientation']['value'] = '\\'.join(elements['patient_orientation']['value'])



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="metadata_to_db",
    version="0.0.7",
    author="Matt-Conrad",
    author_email="mattgrayconrad@gmail.com",
    description="Library for storing image metadata from a directory of images to a PostgreSQL DB.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Matt-Conrad/DicomToDatabase",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


"""Module contains class for handling config files of the INI format."""
from configparser import ConfigParser
import logging
import atexit
import os

class ConfigHandler:
    def __init__(self, configFilename):
        self.configFilename = configFilename

        self.parser = ConfigParser()
        self.readConfigFile()
        atexit.register(self.writeConfigFile)   

    # Functions for external use
    def getConfigFilename(self):
        return self.configFilename

    # Functions for internal use
    def readConfigFile(self):
        logging.info("Reading config file: %s", self.getConfigFilename())
        if os.path.exists(self.configFilename):
            self.parser.read(self.configFilename)
            logging.info("Config file read")
        else:
            raise OSError('File {0} not found'.format(self.getConfigFilename()))

    def writeConfigFile(self):
        logging.info("Writing config file: %s", self.getConfigFilename())
        with open(self.configFilename, 'w') as configfile:
            self.parser.write(configfile)

    def getSection(self, sectionName):
        section = {}
        if self.parser.has_section(sectionName):
            params = self.parser.items(sectionName)
            for param in params:
                section[param[0]] = param[1]
        else:
            raise Exception('Section {0} not found in the {1} file'.format(sectionName, self.configFilename))
        return section
    
    def getSetting(self, sectionName, settingName):
        return self.getSection(sectionName)[settingName]

    def setSetting(self, sectionName, settingName, value):
        self.parser[sectionName][settingName] = value
        self.writeConfigFile()




"""Contains script that moves all DCM tag-values from a directory of DCMs into a PostgreSQL DB."""
import logging
import sys
import os
import json
from pathlib import Path
import psycopg2
import nibabel as nib
import numpy as np
# This line is so modules using this package as a submodule can use this.
sys.path.append(os.path.dirname(os.path.abspath(__file__)).replace('\\', '/'))
#
from config import config
import basic_db_ops as bdo

# Types of files we want from the dataset
DESIRED_SUFFIXES = ['injured.nii', 'uninjured.nii']

def nifti_to_db(elements_json, config_file_name, sectionName):
    """Move all desired NIFTI metadata from a directory full of NIFTIs into a PostgreSQL DB.

    This function goes through all of the NIFTI files in the directory specified in the 'postgresql'
    section of the config file specified by config_file_name. For each NIFTI, the function reads the
    values of the tags specified in the JSON named in elements_json and stores them all in a
    PostgreSQL DB table where the columns are the tags and each row is a NIFTI.

    Parameters
    ----------
    elements_json : string
        The name of the JSON that contains the list of elements we want to read from the DICOM
    config_file_name : string
        The name of the file that contains DB and DICOM folder info
    sectionName : string
        Name of the section in the elements_json that has the column info for that table
    """
    logging.info('Attempting to store NIFTI metadata from DCMs in a folder to Postgres DB')
    
    # Create the database if it isn't already there
    dbName = config(filename=config_file_name, section='postgresql')['database']
    if not bdo.dbExists(config_file_name, dbName):
        bdo.createNewDb(dbName)

    # Create table if it isn't already there
    dbName = config(filename=config_file_name, section='postgresql')['database']
    tableName = config(filename=config_file_name, section='table_info')['metadata_tableName']
    if not bdo.tableExists(config_file_name, dbName, tableName):
        bdo.addTableToDb(tableName, elements_json, config_file_name, sectionName)

    # Open the json with the list of elements we're interested in
    with open(elements_json) as fileReader:
        elementsDict = json.load(fileReader)
    elementsOriginal = elementsDict[sectionName]

    folder_path = config(filename=config_file_name, section='nifti_folder')['folder_path']
    pathlist = Path(folder_path).glob('**/*.nii')
    for path in pathlist:
        # read each image in the subdirectories
        file_path = str(path)

        # Only want image files that contain suffixes in the DESIRED_SUFFIXES list
        suffix_exists_in_path = [suffix for suffix in DESIRED_SUFFIXES if(suffix in file_path)] 
        if not bool(suffix_exists_in_path):
            continue

        elements = elementsOriginal.copy()

        logging.info('Storing: ' + file_path)

        # Insert the DICOM metadata as a new record in the Postgres DB
        conn = None
        try:
            # read the connection parameters
            params = config(filename=config_file_name, section='postgresql')
            tableName = config(filename=config_file_name, section='table_info')['metadata_tableName']
            # # connect to the PostgreSQL server
            conn = psycopg2.connect(**params)
            cur = conn.cursor()
            # Create the SQL query to be used
            sqlQuery, values = createSqlQuery(tableName, elements, file_path)
            logging.debug('SQL Query: %s', sqlQuery)
            # create table one by one
            cur.execute(sqlQuery, values)
            # close communication with the PostgreSQL database server
            cur.close()
            # commit the changes
            conn.commit()
            logging.info('Stored')
        except (psycopg2.DatabaseError) as error:
            logging.warning(error)
            logging.warning('Not stored')
        finally:
            if conn is not None:
                conn.close()

# TODO: Create a more specific name for this function
def createSqlQuery(tableName, elements, file_path):
    """Create the SQL query for inserting a record.

    This function reads each element in elementsDict from the NIFTI specified by file_path and
    formats the data into a SQL query to be later executed by psycopg2.

    Parameters
    ----------
    tableName : string
        The name of the table that the SQL query is to be aimed at
    elements : dict
        The dictionary containing all of the info from the elements_json
    file_path : [type]
        The file path of the DCM we're interested in

    Returns
    -------
    (string, list)
        Return the completed SQL query and a list of the values that will be formatted
        into the SQL query by the psycopg2 execute function.
    """
    img = nib.load(file_path)
    # Go through the list of elements and try to read the value
    for elementName in elements.keys():
        try:
            value = img.header[elementName]
            if isinstance(value, np.ndarray):
                value = value.tolist()
            elements[elementName]['value'] = value
        except (KeyError) as tag: # if the value isn't there, then set it as None
            logging.warning('Cannot read the following NIFTI tag: ' + elementName)
            elements[elementName]['value'] = None
            continue

    # Adjust the data as necessary before storing
    # dataAdjustments(elements)

    # Create the list of values that we're going to use to build the query
    names = ['file_path']
    values = [file_path]
    placeholders = ['%s']
    # Append any value to this list that isn't birth_date or study_date
    for elementName in elements.keys():
        if not elements[elementName]['calculation_only']:
            names.append(elementName)
            values.append(elements[elementName]['value'])
            placeholders.append('%s')

    # Build the SQL query
    sqlQuery = 'INSERT INTO ' + tableName + ' (' + ', '.join(names) + ')' + os.linesep \
        + 'VALUES (' + ', '.join(placeholders) + ');'

    return (sqlQuery, values)

if __name__ == "__main__":
    logging.basicConfig(filename='nifti_to_db.log', level=logging.INFO)
    nifti_to_db('elements.json', 'config.ini', 'nifti_elements')
